#include "header\GBA.h"
#include "header\BG.h"
#include "header\Timer.h"
#include "header\DMA.h"
#include "header\IRQ.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\Blend.h"

#include "lib.h"
#include "iwram.h"
#include "irq.h"
#include "syscall.h"

#include "data.h"
#include "resource\BitMapFont.h"

#include <stdlib.h>
#include <string.h>

//const int __gba_multiboot;  //GBANp

#define BLOCK_CX 8
#define BLOCK_CY 8

#define MAP_X    0
#define MAP_Y    0
#define MAP_CX   240
#define MAP_CY   160

#define MAP_BLOCK_CX  (MAP_CX / BLOCK_CX - 1)
#define MAP_BLOCK_CY  (MAP_CY / BLOCK_CY - 1)
#define MAP_BLOCK_CNT (MAP_BLOCK_CX + MAP_BLOCK_CY)

#define ATTRIBUTE_BLOCK 1
#define ATTRIBUTE_BLANK 2

#define COLOR_BLOCK RGB(31,31,31)

//---------------------------------------------------------------------------
typedef struct {
	u8  cx;
	u8  cy;
	u8  MapX;
	u8  MapY;
	u16 SprNo;
	u16 SprTile;
}ST_PLAYER;

ST_PLAYER PLAYER;

u8 Map[MAP_BLOCK_CX][MAP_BLOCK_CY];
u8 EndFlag;
//---------------------------------------------------------------------------
void Init();

void Input();

void DispChr();
void DispMap();
void DispMapSolve();

void MakeMap();
void MakeWall(s16 x, s16 y);

u8 GetAroundBlockCnt(s16 x, s16 y);
//---------------------------------------------------------------------------
int main(void){
	Mode3SetBitMapFont((u8*)BitMapFont, 8,8,0);
	Mode3DrawBitMapStr(1,1, "Push A Button");
	SetMode(MODE_3 | BG2_ENABLE | OBJ_ENABLE | OBJ_MAP_1D);

	WaitKey(KEY_A);		//SrandĂ܂

	Init();
	MakeMap();
	DispMap();
	DispChr();

	for(;;){
		Input();
		DispChr();

		if(EndFlag){
			DispMapSolve();	//@̕\
			break;
		}
	}

	for(;;){}
	return 0;
}
//---------------------------------------------------------------------------
void Init(){
	EndFlag = 0;
	
	PLAYER.cx      = 8;
	PLAYER.cy      = 8;
	PLAYER.MapX    = 1;
	PLAYER.MapY    = 1;
	PLAYER.SprNo   = 0;
	PLAYER.SprTile = 0;

	u8 x, y;
	for(y=0; y<MAP_BLOCK_CY; y++){
		for(x=0; x<MAP_BLOCK_CX; x++){
			Map[x][y] = ATTRIBUTE_BLANK;
		}
	}

	//Oǂ̍쐬
	//x
	for(x=0; x<MAP_BLOCK_CX; x++){
		Map[x][0             ] = ATTRIBUTE_BLOCK;
		Map[x][MAP_BLOCK_CY-1] = ATTRIBUTE_BLOCK;
	}
	//y
	for(y=0; y<MAP_BLOCK_CY; y++){
		Map[0             ][y] = ATTRIBUTE_BLOCK;
		Map[MAP_BLOCK_CX-1][y] = ATTRIBUTE_BLOCK;
	}

	//Sprite
	SpriteInitialize();
	SpriteSetPal16((u16*)EtfsGetDataOffset(_SPRCHR_IMG_PAL_OFFSET));
	SpriteSetData((u16*)EtfsGetDataOffset(_SPRCHR_IMG_BIN_OFFSET), _SPRCHR_IMG_BIN_SIZE);

	SpriteSetSize(PLAYER.SprNo, SP_SIZE_8, SP_SQUARE, SP_COLOR_16);
	SpriteSetChr(PLAYER.SprNo, 512);

	//ʔwi
	Mode3DrawBox(0, 0, 240, 160, RGB(12,10,10));
}
//---------------------------------------------------------------------------
void DispChr(){
	SpriteMove(PLAYER.SprNo, PLAYER.MapX*BLOCK_CX, PLAYER.MapY*BLOCK_CY);
}
//---------------------------------------------------------------------------
void Input(){
	u16 key;
	u8 tmpPX, tmpPY;

ReInput:

	tmpPX = PLAYER.MapX;
	tmpPY = PLAYER.MapY;

	for(;;){
		key = *KEYS ^ 0x3ff;

		if(key & KEY_UP   ){ tmpPY--;   break; }
		if(key & KEY_DOWN ){ tmpPY++;   break; }
		if(key & KEY_LEFT ){ tmpPX--;   break; }
		if(key & KEY_RIGHT){ tmpPX++;   break; }
		if(key & KEY_B)    { EndFlag=1; break; }
	}//for(;;)

	if(Map[tmpPX][tmpPY] == ATTRIBUTE_BLOCK)
		goto ReInput;

	PLAYER.MapX = tmpPX;
	PLAYER.MapY = tmpPY;

	if(PLAYER.MapX == MAP_BLOCK_CX-2 &&
	   PLAYER.MapY == MAP_BLOCK_CY-2){
		EndFlag = 1;
	}

	while(*KEYS != 0x3ff){}		//͂ɂȂ܂ő҂
}

//---------------------------------------------------------------------------
void DispMap(){
	u8 print_x, print_y;
	u8 x, y;

	for(y=0; y<MAP_BLOCK_CY; y++){
		for(x=0; x<MAP_BLOCK_CX; x++){

			if(Map[x][y] == ATTRIBUTE_BLOCK){
				print_x = (x * BLOCK_CX) + MAP_X;
				print_y = (y * BLOCK_CY) + MAP_Y;

				Mode3DrawBox(print_x, print_y, BLOCK_CX, BLOCK_CY, COLOR_BLOCK);
			}
		}
	}
}
//---------------------------------------------------------------------------
void MakeMap(){
	u8 x,y;
	u8 i;

	for(i=0; i<5; i++){

		for(;;){
			x = rand() % MAP_BLOCK_CX;
			y = rand() % MAP_BLOCK_CY;

			if(x%2) x++;
			if(y%2) y++;

			if(Map[x][y] == ATTRIBUTE_BLOCK){
				break;
			}
		}
		MakeWall((s16)x, (s16)y);
	}

	for(y=0; y<MAP_BLOCK_CY; y+=2){
		for(x=0; x<MAP_BLOCK_CX; x+=2){
			if(Map[x][y] == ATTRIBUTE_BLOCK){
				MakeWall((s16)x, (s16)y);
			}
		}
	}
}
//---------------------------------------------------------------------------
//_ŕǂ̂΂
void MakeWall(s16 x, s16 y){
	s16  tmpX,  tmpY;
	s16 vectX, vectY;
	s8 i, dir;

	for(i=0; i<5; i++){
		tmpX = x;
		tmpY = y;

		dir = rand() % 4;     
		vectX = dir-2+(dir==0)*2;
		vectY = dir-1-(dir==3)*2;

		//1}X
		tmpX += vectX;
		tmpY += vectY;
		if(GetAroundBlockCnt(tmpX, tmpY) <= 3){

			//Q}X
			tmpX += vectX;
			tmpY += vectY;
			if(GetAroundBlockCnt(tmpX, tmpY) == 0){

				Map[tmpX      ][tmpY      ] = ATTRIBUTE_BLOCK;
				Map[tmpX-vectX][tmpY-vectY] = ATTRIBUTE_BLOCK;
				x = tmpX;
				y = tmpY;
				i = -1;

				DispMap();
			}
		}
	}
}
//---------------------------------------------------------------------------
u8 GetAroundBlockCnt(s16 x, s16 y){
	s16 vect[9][2] = { {-1,-1}, { 0,-1}, { 1,-1},
					   {-1, 0}, { 0, 0}, { 1, 0},
					   {-1, 1}, { 0, 1}, { 1, 1},
					 };
	s16 tmpX, tmpY;
	u8 cnt;		//ǂ͈͊ÕJEg
	u8 i;

	cnt=0;
	for(i=0; i<9; i++){
		tmpX = x + vect[i][0];
		tmpY = y + vect[i][1];

		if(tmpX<0 || tmpX>=MAP_BLOCK_CX || tmpY<0 || tmpY>=MAP_BLOCK_CY){
			cnt++;
		}else if(Map[tmpX][tmpY] == ATTRIBUTE_BLOCK){
			cnt++;
		}
	}
	return cnt;
}
//---------------------------------------------------------------------------
void DispMapSolve(){
	u8 x, y;
	u8 flag=1;

	while(flag) {
		flag=0;

		for(y=0; y<MAP_BLOCK_CY; y++){
			for(x=0; x<MAP_BLOCK_CX; x++){

				//X^[gƃS[̈ʒu͏Ȃ
				if( !((x==1 && y==1) || (x==MAP_BLOCK_CX-2 && y==MAP_BLOCK_CY-2)) ){

					if(Map[x][y] == ATTRIBUTE_BLANK){
						if(GetAroundBlockCnt((s16)x, (s16)y) == 7){
							Map[x+0][y+0] = ATTRIBUTE_BLOCK;
							Map[x+0][y-1] = ATTRIBUTE_BLOCK;
							Map[x+0][y+1] = ATTRIBUTE_BLOCK;
							Map[x-1][y+0] = ATTRIBUTE_BLOCK;
							Map[x+1][y+0] = ATTRIBUTE_BLOCK;

							flag = 1;
						}
					}
				}
			}
		}
	}
	DispMap();
}
